//---------------------------------------------------------------------------

#include <vcl.h>

#include <cstdlib>
#include <string>
#include <vector>

#pragma hdrstop

#include <fstream>

#include "MainFormUnit.h"
#include "binfile.hpp"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TFrmMain *FrmMain;


const unsigned char modulesig[] = {0x00, 0xFF, 'C','A','S','I','O', 0x03};

struct modeinfo_t
{
    unsigned char   name[16];
    BOOL            replacable;
};

const modeinfo_t sysmodes[] = {
/*  0 */ {"SYSTEM", 0},
/*  1 */ {"CONTACTS", 1},
/*  2 */ {"MEMO", 1},
/*  3 */ {"SCHEDULE", 1},
/*  4 */ {"CLOCK", 0},
/*  5 */ {"", 0},
/*  6 */ {"Pocket Sheet", 1},
/*  7 */ {"COMMUNICATION", 0},
/*  8 */ {"", 1},
/*  9 */ {"QUICKMEMO", 1},
/* 10 */ {"GAME", 1},
/* 11 */ {"", 1},
/* 12 */ {"EXPENSE", 1},
/* 13 */ {"CONVERSION", 1},
/* 14 */ {"MENU", 1},
/* 15 */ {"MAIL", 1},
};


struct addinheader_t
{
  unsigned char addinsig[8]; /* = {0x00, 0xFF, 'C','A','S','I','O', 0x03}; */
  char model[4]; /* = {'Z','4','8','6'}; */
  char headerversion[4]; /* = {'0','1','0','0'}; */
  unsigned short status;
  unsigned short mode;
  char addinname[16];
  unsigned long addinlength;
  char compiledate[8]; /* e.g. "20020219" stands for Feb. 19th, 2002 */
  char compiletime[4]; /* e.g. "1246" is 46 minutes past 12 o'clock */
  char version[4]; /* "0120" stands for version 1.20 */
  char libdate[8]; /* e.g. "20000215" stands for Feb. 15th, 2000 */
  char libtime[4]; /* e.g. "0940" is 40 minutes past 9 o'clock */
  char libver[4]; /* e.g. "0100" stands for version 1.00 */
  unsigned long offsicon;
  unsigned long offslicon;
  char comment[64];
};

struct module_t
{
	DWORD 	  address;
	char      name[16];
	WORD      mode;
	DWORD     size;
	DWORD     maxsize;
	BOOL      replacable;

    module_t (void) {}
    module_t (DWORD addr, WORD md, DWORD sz)
     : address (addr), mode (md), size (sz), replacable (0)
    {
    }
};

const DWORD     RADIX_BINARY         =  2;
const DWORD     RADIX_OCTAL          =  8;
const DWORD     RADIX_DECIMAL        = 10;
const DWORD     RADIX_HEXADECIMAL    = 16;

AnsiString dw2str_format (DWORD dw, DWORD radix, DWORD min_cnt)
{
    DWORD len;
    char buf[256 + 34];
    char* ptr = buf + 256;

    if (min_cnt > (256 + 33))
        return ("");
    std::itoa (dw, ptr, radix);

    len = strlen (ptr);
    while (min_cnt > len++)
        *--ptr = '0';
    return (AnsiString (ptr).UpperCase ());
}

std::vector<module_t> modules;
ucl::BinFile bfile;
/*char* bfile_ptr = NULL;
DWORD bfile_size;*/


//---------------------------------------------------------------------------
__fastcall TFrmMain::TFrmMain(TComponent* Owner)
    : TForm(Owner)
{
}

//---------------------------------------------------------------------------
void __fastcall TFrmMain::LbxModulesDrawItem(TWinControl *Control, int Index,
      TRect &Rect, TOwnerDrawState State)
{
    int x1, x2, x3, x4, x5;
    DWORD radix = ChkHexDraw->Checked ? RADIX_HEXADECIMAL : RADIX_DECIMAL;
    DWORD numbers = ChkHexDraw->Checked ? 8 : 0;

    x1 = Rect.left;
    x2 = x1 + HctModules->Sections->Items[0]->Width;
    x3 = x2 + HctModules->Sections->Items[1]->Width;
    x4 = x3 + HctModules->Sections->Items[2]->Width;
    x5 = x4 + HctModules->Sections->Items[3]->Width;

	LbxModules->Canvas->Brush->Color = State.Contains (odSelected)
											? clHighlight
											: clWindow;
    LbxModules->Canvas->FillRect (Rect);

    LbxModules->Canvas->Font->Color = modules[Index].replacable
											? (State.Contains (odSelected)
												? clHighlightText
												: clWindowText)
											: clGrayText;

    LbxModules->Canvas->Font->Name = "Courier New";
    LbxModules->Canvas->TextOut (x1, Rect.Top,
        dw2str_format (modules[Index].address, RADIX_HEXADECIMAL, 8));
    LbxModules->Canvas->TextOut (x3, Rect.Top,
        dw2str_format (modules[Index].mode, RADIX_HEXADECIMAL, 4));
    LbxModules->Canvas->TextOut (x4, Rect.Top,
        dw2str_format (modules[Index].size, radix, numbers));
    LbxModules->Canvas->TextOut (x5, Rect.Top,
        dw2str_format (modules[Index].maxsize, radix, numbers));

    LbxModules->Canvas->Font->Name = LbxModules->Font->Name;
    LbxModules->Canvas->TextOut (x2, Rect.Top,
        modules[Index].name);
}
//---------------------------------------------------------------------------


void __fastcall TFrmMain::scanForModules (void)
{
    // scan file for modules
    int offs;
    addinheader_t* aih;
    BYTE* pcur = &bfile.data[0];
    BYTE* plast = &bfile.data[0];
    BYTE* pend = &bfile.data[0] + bfile.data.size ();
    unsigned char* psch;
    unsigned char* pdch;

    modules.clear ();
    LbxModules->Items->Clear ();
	BtnReplace->Enabled = false;
	BtnExport->Enabled = false;

    for (; pcur < pend; pcur += 1024)
	{
		if ((pend - pcur) < sizeof (addinheader_t))
			break;
        aih = reinterpret_cast <addinheader_t*> (pcur);
		if (memcmp (aih->addinsig, modulesig, 8) == 0)
        { /* module found */
			offs = modules.size ();
            modules.push_back (module_t (pcur - &bfile.data[0], aih->mode, aih->addinlength));
            psch = aih->addinname;
			pdch = modules[offs].name;

				/* copy name */
            while ((*psch != 0x00) && (*psch != 0xFF) && (psch < (aih->addinname + 15)))
				*pdch++ = *psch++;
            *pdch = 0x00;
            if ((aih->mode >> 8) < 0x10)
			{	/* is system application */
                modules[offs].replacable = ChkExperimental->Checked || sysmodes[aih->mode >> 8].replacable;
				if (pdch == modules[offs].name) /* if no name is given, use std sysapp name */
                    strcpy (modules[offs].name, sysmodes[aih->mode >> 8].name);
            }

			if ((plast != pcur) && offs)
				modules[offs - 1].maxsize = pcur - plast;
            plast = pcur;
            LbxModules->Items->Add ("");
        }
    }
	if (modules.size () > 0)
	{
		modules[modules.size () - 1].maxsize = pend - plast;
		BtnSave->Enabled = true;
		BtnSaveHex->Enabled = true;
	}

    BtnReplace->Enabled = false;
    LbxModules->ItemIndex = -1;
}

void __fastcall TFrmMain::BtnOSFileClick(TObject *Sender)
{
	/*std::fstream bfile;
    DWORD filesize, filesize_high;
    HANDLE hFile;*/

    if (!OpdOSFile->Execute ())
        return;

    if (ExtractFileExt (OpdOSFile->FileName).LowerCase () == ".hex")
        bfile.loadFromHexFile (OpdOSFile->FileName.c_str ());
    else
        bfile.loadFromFile (OpdOSFile->FileName.c_str ());
    /*hFile = CreateFile (OpdOSFile->FileName.c_str (),
      FILE_READ_DATA, FILE_SHARE_READ, NULL, OPEN_EXISTING,
      FILE_ATTRIBUTE_NORMAL, NULL);
    filesize = GetFileSize (hFile, &filesize_high);
    CloseHandle (hFile);

	bfile.open (OpdOSFile->FileName.c_str (), ios::in | ios::binary);

    if (&bfile.data[0])
        delete[] bfile_ptr;
	bfile_ptr = new char[filesize];
    bfile.read (bfile_ptr, filesize);

    bfile_size = filesize;

    bfile.close ();*/

    EdtOSFile->Text = OpdOSFile->FileName;
    scanForModules ();
}
//---------------------------------------------------------------------------
void __fastcall TFrmMain::FormClose(TObject *Sender, TCloseAction &Action)
{
	/*if (bfile_ptr)
	{
		delete[] bfile_ptr;
		bfile_ptr = NULL;
	}*/
}
//---------------------------------------------------------------------------
void __fastcall TFrmMain::BtnExitClick(TObject *Sender)
{
    Close ();
}
//---------------------------------------------------------------------------
void __fastcall TFrmMain::LbxModulesClick(TObject *Sender)
{
	module_t* module = &modules[LbxModules->ItemIndex];
	bool is_system_module = !(module->mode & 0xFF00) && (module->mode != 0x8000);
	bool is_at_64k_border = !(module->address & 0x0000FFFF);

	BtnReplace->Enabled = module->replacable && !is_system_module;
	BtnExport->Enabled = is_at_64k_border && !is_system_module;
}
//---------------------------------------------------------------------------
void __fastcall TFrmMain::ModulesRefresh(TObject *Sender)
{
	scanForModules ();
    LbxModules->Refresh ();
}
//---------------------------------------------------------------------------
void __fastcall TFrmMain::BtnReplaceClick(TObject *Sender)
{
	std::fstream sfile;
	DWORD filesize, filesize_high;
	HANDLE hFile;
	BYTE headersig[8];
	module_t* module = &modules[LbxModules->ItemIndex];
	BYTE* pb = &bfile.data[0] + module->address;
	BYTE* plast = pb + module->maxsize;
	addinheader_t* aih = reinterpret_cast <addinheader_t*> (pb);
	WORD mode = aih->mode;

	if (module->address & 0x0000FFFF)
			Application->MessageBox (("CARE:\r"
				"Applications placed here have to be linked with\n"
				"special linker settings:\n\n"
				"1. In the MAKEFILE, change the line\n"
				"     include ..\\COM_LNK\\MakeSDK.2\n"
				"   to\n"
				"     include MakeSDK.2\n"
				"2. Copy ..\\COM_LNK\\MakeSDK.2 to the project folder\n"
				"3. In the local MakeSDK.2, change the line\n"
				"     TARSEG = -TTEXT 80000\n"
				"   to\n"
				"     TARSEG = -TTEXT 8"
				  + dw2str_format (module->address & 0x0000FFFF,
								   RADIX_HEXADECIMAL, 4)
				  + "\n\n" +
				"That is necessary since the selected application\n"
				"is not located at a 64kB border.").c_str (),	
				"Caution", MB_ICONWARNING);

	if (!OpdReplaceFile->Execute ())
			return;

	hFile = CreateFile (OpdReplaceFile->FileName.c_str (),
	  FILE_READ_DATA, FILE_SHARE_READ, NULL, OPEN_EXISTING,
	  FILE_ATTRIBUTE_NORMAL, NULL);
	filesize = GetFileSize (hFile, &filesize_high);
	CloseHandle (hFile);

	sfile.open (OpdReplaceFile->FileName.c_str (), ios::in | ios::binary);
	sfile.seekg (0, ios::beg);
	sfile.read (headersig, 8);
	if (memcmp (headersig, modulesig, 8))
	{
		Application->MessageBox ("Invalid file format", "Error", MB_ICONERROR);
		sfile.close ();
		return;
	}
	if (filesize > module->maxsize)
	{
		Application->MessageBox ("File exceeds maximal size", "Error", MB_ICONERROR);
		sfile.close ();
		return;
	}

	sfile.read (pb + 8, filesize);
	aih->mode = mode;

	for (pb += filesize; pb < plast; *pb++ = 0xFF)
		;	/* fill remaining space with 0xFF */

	sfile.close ();

	scanForModules ();
	LbxModules->Refresh ();
}

AnsiString str_ext (const AnsiString& filename, const AnsiString& ext)
{
	if (!filename.LowerCase ().Pos (ext.LowerCase ()))
		if (filename.SubString (filename.Length () - ext.Length () + 1, ext.Length ()).LowerCase () != ext.LowerCase ())
			return (filename + ext);
	return (filename);
}
//---------------------------------------------------------------------------
void __fastcall TFrmMain::BtnSaveClick(TObject *Sender)
{
	//std::fstream bfile;

	if (!SvdOSFile->Execute ())
		return;
	
	//EdtOSFile->Text = str_ext (SvdOSFile->FileName, ".bin").c_str ();

    bfile.saveToFile (str_ext (SvdOSFile->FileName, ".bin").c_str ());
	/*bfile.open (str_ext (SvdOSFile->FileName, ".bin").c_str (), ios::out | ios::binary);
	bfile.write (&bfile.data[0], bfile_size);
	bfile.close ();*/
}
//---------------------------------------------------------------------------
void __fastcall TFrmMain::BtnExportClick(TObject *Sender)
{
	std::fstream sfile;

	if (!SvdAppFile->Execute ())
		return;

	sfile.open (str_ext (SvdAppFile->FileName, ".bin").c_str (), ios::out | ios::binary);
	sfile.write (&bfile.data[0] + modules[LbxModules->ItemIndex].address, modules[LbxModules->ItemIndex].size);
	sfile.close ();
}
//---------------------------------------------------------------------------

void __fastcall TFrmMain::BtnAboutClick(TObject *Sender)
{
	Application->MessageBox ("PVOSEdit v1.10\n\n"
	  "(c) 2006 by AUDACIA Software\n"
	  "www.audacia-software.de\n\n"
	  "CAUTION: YOU USE THIS SOFTWARE ON YOUR OWN RISK!\n"
	  "ALWAYS TEST RESULT IN SIMULATOR!",
	  "About PVOSEdit", MB_ICONINFORMATION);
}
//---------------------------------------------------------------------------


void __fastcall TFrmMain::BtnSaveHexClick(TObject *Sender)
{
	if (!SvdOSHexFile->Execute ())
		return;
	
    bfile.saveToHexFile (str_ext (SvdOSHexFile->FileName, ".hex").c_str ());
}
//---------------------------------------------------------------------------

